import openai
import openai_async
import os
import sys

import textwrap

def word_wrap_except_code_blocks(text):
    paragraphs = text.split('\n\n') # Split text into paragraphs
    wrapped_paragraphs = []
    in_code_block = False

    for paragraph in paragraphs:
        if paragraph.startswith('```') and paragraph.endswith('```'):
            # If this is a code block, skip wrapping and add it as is
            wrapped_paragraphs.append(paragraph)
            continue

        if paragraph.startswith('```'):
            # Beginning of a code block
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue

        if paragraph.endswith('```'):
            # End of a code block
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue

        if in_code_block:
            # If in a code block, add paragraph as is
            wrapped_paragraphs.append(paragraph)
        else:
            # If not in a code block, wrap paragraph and add to list
            wrapped_paragraph = textwrap.fill(paragraph) # , width=80)
            wrapped_paragraphs.append(wrapped_paragraph)

    # Join all paragraphs into a single string
    wrapped_text = '\n\n'.join(wrapped_paragraphs)

    return wrapped_text


async def why(self, arg):
    user_prompt = 'Explain what the root cause of this error is, given the following source code and traceback, and propose a fix.'
    user_prompt += '\n'
    user_prompt += 'source code:\n```\n'

    stack_trace = ''
    
    stack_frames = len(self.stack)
    try:
        import sys
        exception_name = sys.exc_info()[0].__name__
        exception_value = sys.exc_info()[1]
    except:
        print("The command 'why' only works when there is an uncaught exception. Try running 'python3 -m chatdbg -c continue'.")
        return

    # print(dir(self))
    for frame_lineno in self.stack:
        import inspect
        frame, lineno = frame_lineno
        if not frame.f_code.co_filename.startswith(os.getcwd()):
            stack_frames -= 1
            continue
        positions = inspect.getframeinfo(frame).positions
        try:
            # user_prompt += '#' + '-' * 60 + '\n'
            lines = inspect.getsourcelines(frame)[0]
            for index, line in enumerate(lines, frame.f_code.co_firstlineno):
                user_prompt += '  '
                user_prompt += line.rstrip() + '\n'
                if index == lineno:
                    leading_spaces = len(line) - len(line.lstrip())
                    stack_trace += f'{stack_frames}: ' + line.strip() + '\n'
                    stack_trace += ' ' * len(str(stack_frames)) + '  ' + ' ' * (positions.col_offset - leading_spaces) + '^' * (positions.end_col_offset - positions.col_offset) + '\n'
                if index >= lineno:
                    break
        except:
            pass
        stack_frames -= 1
    user_prompt += '```\n'
    user_prompt += 'stack trace:\n'
    user_prompt += f'```\n{stack_trace}```\n'
    user_prompt += f'Exception: {exception_name} ({exception_value})\n'

    # print(user_prompt)
    import httpx
    text = ''
    try:
        completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload={'model': 'gpt-3.5-turbo', 'messages': [{'role': 'user', 'content': user_prompt}]})
        json_payload = completion.json()
        text = json_payload['choices'][0]['message']['content']
    except (openai.error.AuthenticationError, httpx.LocalProtocolError):
        print()
        print('You need an OpenAI key to use commentator. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        import sys
        sys.exit(1)
    except Exception as e:
        print(f"EXCEPTION {e}")
        pass
    print(word_wrap_except_code_blocks(text))



import openai
import openai_async
import os
import sys
import textwrap

def word_wrap_except_code_blocks(text: str) -> str:
    """Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split('\n\n')
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # If this paragraph starts and ends with a code block, add it as is.
        if paragraph.startswith('```') and paragraph.endswith('```'):
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the beginning of a code block add it as is.
        if paragraph.startswith('```'):
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the end of a code block stop skipping text.
        if paragraph.endswith('```'):
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue
        # If we are currently in a code block add the paragraph as is.
        if in_code_block:
            wrapped_paragraphs.append(paragraph)
        else:
            # Otherwise, apply text wrapping to the paragraph.
            wrapped_paragraph = textwrap.fill(paragraph)
            wrapped_paragraphs.append(wrapped_paragraph)
    # Join all paragraphs into a single string
    wrapped_text = '\n\n'.join(wrapped_paragraphs)
    return wrapped_text

def read_lines(file_path: str, start_line: int, end_line: int) -> str:
    """
    Read lines from a file and return a string containing the lines between start_line and end_line.

    Args:
        file_path (str): The path of the file to read.
        start_line (int): The line number of the first line to include (1-indexed).
        end_line (int): The line number of the last line to include.

    Returns:
        str: A string containing the lines between start_line and end_line.

    """
    # open the file for reading
    with open(file_path, 'r') as f:
        # read all the lines from the file
        lines = f.readlines()
        # remove trailing newline characters
        lines = [line.rstrip() for line in lines]
    # convert start_line to 0-based indexing
    start_line = max(0, start_line - 1)
    # ensure end_line is within range
    end_line = min(len(lines), end_line)
    # return the requested lines as a string
    return '\n'.join(lines[start_line:end_line])

async def explain(source_code: str, traceback: str, exception: str) -> None:
    import httpx
    user_prompt = 'Explain what the root cause of this error is, given the following source code and traceback, and propose a fix.'
    user_prompt += '\n'
    user_prompt += 'source code:\n```\n'
    user_prompt += source_code + '\n```\n'
    user_prompt += traceback + '\n\n'
    user_prompt += 'stop reason = ' + exception + '\n'
    text = ''
    
    # uncomment to view prompt and not send to OpenAI
    #print(user_prompt)
    #return

    if not 'OPENAI_API_KEY' in os.environ:
        print('You need a valid OpenAI key to use ChatDBG. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        return
    
    try:
        completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload={'model': 'gpt-3.5-turbo', 'messages': [{'role': 'user', 'content': user_prompt}]})
        json_payload = completion.json()
        text = json_payload['choices'][0]['message']['content']
    except (openai.error.AuthenticationError, httpx.LocalProtocolError, KeyError):
        # Something went wrong.
        print()
        print('You need a valid OpenAI key to use ChatDBG. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        import sys
        sys.exit(1)
    except Exception as e:
        print(f'EXCEPTION {e}, {type(e)}')
        pass
    print(word_wrap_except_code_blocks(text))




#! /usr/bin/env python3

from . import pdb
from .pdb import Pdb, Restart, _ModuleTarget, _ScriptTarget

import asyncio
import sys
import traceback

from . import chatdbg_why


class ChatDBG(Pdb):
    def do_why(self, arg):
        asyncio.run(chatdbg_why.why(self, arg))


import importlib.metadata

_usage = f"""\
usage: chatdbg [-c command] ... [-m module | pyfile] [arg] ...

A Python debugger that uses AI to tell you `why`.
(version {importlib.metadata.metadata('ChatDBG')['Version']})

https://github.com/plasma-umass/ChatDBG

Debug the Python program given by pyfile. Alternatively,
an executable module or package to debug can be specified using
the -m switch.

Initial commands are read from .pdbrc files in your home directory
and in the current directory, if they exist.  Commands supplied with
-c are executed after commands from .pdbrc files.

To let the script run until an exception occurs, use "-c continue".
You can then type `why` to get an explanation of the root cause of
the exception, along with a suggested fix. NOTE: you must have an
OpenAI key saved as the environment variable OPENAI_API_KEY.
You can get a key here: https://openai.com/api/

To let the script run up to a given line X in the debugged file, use
"-c 'until X'"."""


def main():
    import getopt

    opts, args = getopt.getopt(sys.argv[1:], "mhc:", ["help", "command="])

    if not args:
        print(_usage)
        sys.exit(2)

    if any(opt in ["-h", "--help"] for opt, optarg in opts):
        print(_usage)
        sys.exit()

    commands = [optarg for opt, optarg in opts if opt in ["-c", "--command"]]

    module_indicated = any(opt in ["-m"] for opt, optarg in opts)
    cls = _ModuleTarget if module_indicated else _ScriptTarget
    target = cls(args[0])

    target.check()

    sys.argv[:] = args  # Hide "pdb.py" and pdb options from argument list

    # Note on saving/restoring sys.argv: it's a good idea when sys.argv was
    # modified by the script being debugged. It's a bad idea when it was
    # changed by the user from the command line. There is a "restart" command
    # which allows explicit specification of command line arguments.
    pdb = ChatDBG()
    pdb.rcLines.extend(commands)
    while True:
        try:
            pdb._run(target)
            if pdb._user_requested_quit:
                break
            print("The program finished and will be restarted")
        except Restart:
            print("Restarting", target, "with arguments:")
            print("\t" + " ".join(sys.argv[1:]))
        except SystemExit:
            # In most cases SystemExit does not warrant a post-mortem session.
            print("The program exited via sys.exit(). Exit status:", end=" ")
            print(sys.exc_info()[1])
        except SyntaxError:
            traceback.print_exc()
            sys.exit(1)
        except:
            traceback.print_exc()
            print("Uncaught exception. Entering post mortem debugging")
            print("Running 'cont' or 'step' will restart the program")
            t = sys.exc_info()[2]
            pdb.interaction(None, t)
            print("Post mortem debugger finished. The " + target + " will be restarted")


# When invoked as main program, invoke the debugger on a script
if __name__ == "__main__":
    import chatdbg

    chatdbg.main()



import ast_comments as ast
import openai_async
import asyncio
import logging
import openai
import os
import sys

from rich.progress import Progress

from typing import (
    Any,
    cast,
    Deque,
    DefaultDict,
    Dict,
    FrozenSet,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
import typing

logname = "chatdbg.log"
logging.basicConfig(
    filename=logname,
    filemode="w",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.INFO,
)

logging.info("Running ChatDBG.")


async def get_result(
    system_content, user_content, pbar, progress, model="gpt-3.5-turbo", max_trials=3
) -> Optional[openai.api_resources.Completion]:
    import httpx

    try:
        for trial in range(max_trials):
            completion = await openai_async.chat_complete(
                openai.api_key,
                timeout=30,
                payload={
                    "model": model,
                    "messages": [
                        {
                            "role": "system",
                            "content": system_content,
                            "role": "user",
                            "content": user_content,
                        }
                    ],
                },
            )
            code_block = extract_code_block(completion.json())
            logging.info(f"PROCESSING {code_block}")
    except (openai.error.AuthenticationError, httpx.LocalProtocolError):
        print()
        print(
            "You need an OpenAI key to use commentator. You can get a key here: https://openai.com/api/"
        )
        print(
            "Invoke commentator with the api-key argument or set the environment variable OPENAI_API_KEY."
        )
        import sys

        sys.exit(1)
    except Exception as e:
        return ""
    progress.update(pbar, advance=1)
    return code_block


def find_code_start(code: str) -> int:
    """
    Finds the starting location of a code block in a string.

    Args:
        code: A string containing code.

    Returns:
        An integer representing the starting position of the code block.

    """
    lines = code.split("\n")
    i = 0
    while i < len(lines) and lines[i].strip() == "":
        i += 1
    first_line = lines[i].strip()
    if first_line == "```":
        return 3
    if first_line.startswith("```"):
        word = first_line[3:].strip()
        if len(word) > 0 and " " not in word:
            return len(word) + 3
    return -1


def extract_code_block(completion: dict) -> str:
    """
    Extracts code block from the given completion dictionary.

    Args:
        completion (dict): Completion dictionary containing text and other data.

    Returns:
        str: Extracted code block from the completion dictionary.
    """
    c = completion
    text = c["choices"][0]["message"]["content"]
    first_index = find_code_start(text)
    second_index = text.find("```", first_index + 1)
    if first_index == -1 or second_index == -1:
        code_block = text
    else:
        code_block = text[first_index:second_index]
    return code_block



#!env python3
from typing import Tuple, Union
import lldb
import asyncio

import utils

def __lldb_init_module(debugger: lldb.SBDebugger, internal_dict: dict) -> None:
    # Install the `why` command.
    debugger.HandleCommand('command script add -f chatdbg_lldb.why why')
    # Update the prompt.
    debugger.HandleCommand("settings set prompt '(ChatDBG lldb) '")

    
# From lldbinit

def get_process() -> Union[None, lldb.SBProcess]:
    """
    Get the process that the current target owns.
    :return: An lldb object representing the process (lldb.SBProcess) that this target owns.
    """
    return get_target().process

def get_frame() -> lldb.SBFrame:
    """
    Get the current frame of the running process.

    :return: The current frame of the running process as an SBFrame object.
    """
    # Initialize frame variable to None
    frame = None
    for thread in get_process():
        # Loop through the threads in the process
        if thread.GetStopReason() != lldb.eStopReasonNone and thread.GetStopReason() != lldb.eStopReasonInvalid:
            # If the stop reason is not "none" or "invalid", get the frame at index 0 and break the loop.
            frame = thread.GetFrameAtIndex(0)
            break
    if not frame:
        # If frame is None, print a warning message
        # print("[-] warning: get_frame() failed. Is the target binary started?")
        # The warning message has been commented out, so just pass.
        pass
    # Return the current frame.
    return frame

def get_thread() -> lldb.SBThread:
    """
    Returns the currently stopped thread in the debugged process.
    :return: The currently stopped thread or None if no thread is stopped.
    """
    thread = None
    # Iterate over threads in the process
    for _thread in get_process():
        # Check if thread is stopped for a valid reason
        if _thread.GetStopReason() != lldb.eStopReasonNone and _thread.GetStopReason() != lldb.eStopReasonInvalid:
            thread = _thread
    if not thread:
        # No stopped thread was found
        pass
    return thread

def get_target() -> lldb.SBTarget:
    target = lldb.debugger.GetSelectedTarget()
    if not target:
        return None
    return target
#

import re

def get_function_by_line(source_code: str, line_number: int) -> str:
    source_code = re.sub('\\/\\/.*', '', source_code)
    source_code = re.sub('\\/\\*[\\s\\S]*?\\*\\/', '', source_code)
    source_code = re.sub('\\s+', '', source_code)
    function_regex = '(?P<signature>(static|const)?[\\w\\s\\*]+)(?P<name>\\w+)\\((?P<args>[\\w\\s\\*,]*)\\)(?P<body>{[\\s\\S]*?})'
    functions = re.finditer(function_regex, source_code)
    for function in functions:
        function_body = function.group('body')
        line_count = function_body.count('\n')
        print('checking ', function_body)
        if line_number <= line_count:
            return function.group('signature') + function.group('name') + '(' + function.group('args') + ')' + function_body
        line_number -= line_count
    return None

def buildPrompt(debugger: any) -> Tuple[str, str, str]:
    import os
    target = get_target()
    if not target:
        return ''
    thread = get_thread()
    if not thread:
        return ''
    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        return ''
    frame = thread.GetFrameAtIndex(0)
    stack_trace = ''
    source_code = ''
    
    # magic number - don't bother walking up more than this many frames.
    # This is just to prevent overwhelming OpenAI (or to cope with a stack overflow!).
    max_frames = 10
    
    for index, frame in enumerate(thread):
        if index >= max_frames:
            break
        function = frame.GetFunction()
        if not function:
            continue
        full_func_name = frame.GetFunctionName()
        func_name = full_func_name.split('(')[0]
        arg_list = []

        # Build up an array of argument values to the function.
        for i in range(len(frame.GetFunction().GetType().GetFunctionArgumentTypes())):
            arg_list.append(str(frame.FindVariable(frame.GetFunction().GetArgumentName(i))).split('=')[1].strip())
            
        line_entry = frame.GetLineEntry()
        file_spec = line_entry.GetFileSpec()
        file_name = file_spec.GetFilename()
        directory = file_spec.GetDirectory()
        full_file_name = os.path.join(directory, file_name)
        line_num = line_entry.GetLine()
        col_num = line_entry.GetColumn()
        stack_trace += f'frame {index}: {func_name}({",".join(arg_list)}) at {file_name}:{line_num}:{col_num}\n'
        try:
            source_code += f'/* frame {index} */\n'
            source_code += utils.read_lines(full_file_name, line_num - 10, line_num) + '\n'
            source_code += '-' * (col_num - 1) + '^' + '\n\n'
        except:
            # Couldn't find the source for some reason. Skip the file.
            pass
    error_reason = thread.GetStopDescription(255)
    return (source_code, stack_trace, error_reason)

def why(debugger: lldb.SBDebugger, command: str, result: str, internal_dict: dict) -> None:
    """
    Check if program is attached to a debugger.
    Check if code has been run before executing the `why` command.
    Check if execution stopped at a breakpoint or an error.
    Get source code, stack trace, and exception, and call `explain` function using asyncio.
    """
    if not get_target():
        # Check if program is attached to a debugger.
        print('Must be attached to a program to ask `why`.')
        return
    # Get the current thread.
    thread = get_thread()
    if not thread:
        # Check if code has been run before executing the `why` command.
        print('Must run the code first to ask `why`.')
        return
    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        # Check if execution stopped at a breakpoint or an error.
        print('Execution stopped at a breakpoint, not an error.')
        return
    the_prompt = buildPrompt(debugger)
    # Call `explain` function with pieces of the_prompt  as arguments.
    asyncio.run(utils.explain(the_prompt[0], the_prompt[1], the_prompt[2]))



import lldb

def __lldb_init_module(debugger, internal_dict):
    debugger.HandleCommand('command script add -f MyModule.helloWorld hello')

def helloWorld(debugger, command, result, internal_dict):
    print('Hello, world!')


# Add 'source <path to chatdbg>/chatdbg_gdb.py' to ~/.gdbinit

import asyncio
import gdb
import os
import openai
import openai_async
import textwrap

def word_wrap_except_code_blocks(text: str) -> str:
    """Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split('\n\n')
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # If this paragraph starts and ends with a code block, add it as is.
        if paragraph.startswith('```') and paragraph.endswith('```'):
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the beginning of a code block add it as is.
        if paragraph.startswith('```'):
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the end of a code block stop skipping text.
        if paragraph.endswith('```'):
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue
        # If we are currently in a code block add the paragraph as is.
        if in_code_block:
            wrapped_paragraphs.append(paragraph)
        else:
            # Otherwise, apply text wrapping to the paragraph.
            wrapped_paragraph = textwrap.fill(paragraph)
            wrapped_paragraphs.append(wrapped_paragraph)
    # Join all paragraphs into a single string
    wrapped_text = '\n\n'.join(wrapped_paragraphs)
    return wrapped_text

async def explain(source_code: str, traceback: str, exception: str) -> None:
    import httpx
    user_prompt = 'Explain what the root cause of this error is, given the following source code and traceback, and propose a fix.'
    user_prompt += '\n'
    user_prompt += 'source code:\n```\n'
    user_prompt += source_code + '\n```\n'
    user_prompt += traceback + '\n\n'
    user_prompt += 'stop reason = ' + exception + '\n'
    text = ''
    
    # uncomment to view prompt and not send to OpenAI
    #print(user_prompt)
    #return


    try:
        completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload={'model': 'gpt-3.5-turbo', 'messages': [{'role': 'user', 'content': user_prompt}]})
        json_payload = completion.json()
        text = json_payload['choices'][0]['message']['content']
    except (openai.error.AuthenticationError, httpx.LocalProtocolError):
        print()
        print('You need an OpenAI key to use commentator. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        import sys
        sys.exit(1)
    except Exception as e:
        print(f'EXCEPTION {e}')
        pass
    print(word_wrap_except_code_blocks(text))



def read_lines_list(file_path: str, start_line: int, end_line: int) -> [str]:
    """
    Read lines from a file and return a list containing the lines between start_line and end_line.

    Args:
        file_path (str): The path of the file to read.
        start_line (int): The line number of the first line to include (1-indexed).
        end_line (int): The line number of the last line to include.

    Returns:
        [str]: A list of the lines between start_line and end_line.

    """
    # open the file for reading
    with open(file_path, 'r') as f:
        # read all the lines from the file
        lines = f.readlines()
        # remove trailing newline characters
        lines = [line.rstrip() for line in lines]
    # convert start_line to 0-based indexing
    start_line = max(0, start_line - 1)
    # ensure end_line is within range
    end_line = min(len(lines), end_line)
    # return the requested lines as a list
    return lines[start_line:end_line]

# Set the prompt to gdb-ChatDBG
gdb.prompt_hook = lambda x: "(gdb-ChatDBG) "

last_error_type = ""

def stop_handler(event):
    """Sets last error type so we can report it later."""
    # Check if the event is a stop event
    global last_error_type
    if not hasattr(event, 'stop_signal'):
        last_error_type = "" # Not a real error (e.g., a breakpoint)
        return
    if event.stop_signal is not None:
        last_error_type = event.stop_signal

gdb.events.stop.connect(stop_handler)

        
# Implement the command `why`
class Why(gdb.Command):

    def __init__(self):
        gdb.Command.__init__(self, "why", gdb.COMMAND_USER)

    def invoke(self, arg, from_tty):
        try:
            frame = gdb.selected_frame()
        except:
            print("Must run the code first to ask `why`.")
            return
        global last_error_type
        if not last_error_type:
            # Assume we are running from a core dump,
            # which _probably_ means a SEGV.
            last_error_type = 'SIGSEGV'
        the_prompt = buildPrompt()
        #print(the_prompt[0])
        #print(the_prompt[1])
        #print(the_prompt[2])
        # Call `explain` function with pieces of the_prompt  as arguments.
        asyncio.run(explain(the_prompt[0], the_prompt[1], the_prompt[2]))
        
Why()

def buildPrompt() -> str:
    thread = gdb.selected_thread()
    if not thread:
        return ''

    stack_trace = ""
    source_code = ""

    frames = []
    frame = gdb.selected_frame()

    # magic number - don't bother walking up more than this many frames.
    # This is just to prevent overwhelming OpenAI (or to cope with a stack overflow!).
    max_frames = 10
    
    # Walk the stack and build up the frames list.
    while frame is not None and max_frames > 0:
        func_name = frame.name()
        symtab_and_line = frame.find_sal()
        if symtab_and_line.symtab is not None:
            filename = symtab_and_line.symtab.filename
        else:
            filename = None
        if symtab_and_line.line is not None:
            lineno = symtab_and_line.line
            colno = None
        else:
            lineno = None
            colno = None
        args = []
        block = frame.block()
        for symbol in block:
            if symbol.is_argument:
                name = symbol.name
                value = frame.read_var(name)
                args.append((name, value))
        frames.append((filename, func_name, args, lineno, colno))
        frame = frame.older()
        max_frames -= 1

    # Now build the stack trace and source code strings.
    for i, frame_info in enumerate(frames):
        file_name = frame_info[0]
        func_name = frame_info[1]
        line_num = frame_info[3]
        arg_list = []
        for arg in frame_info[2]:
            arg_list.append(str(arg[1])) # Note: arg[0] is the name of the argument
        stack_trace += f'frame {i}: {func_name}({",".join(arg_list)}) at {file_name}:{line_num}\n'
        try:
            source_code += f'/* frame {i} */\n'
            lines = read_lines_list(file_name, line_num - 10, line_num)
            source_code += '\n'.join(lines) + '\n'
            # Get the spaces before the last line.
            num_spaces = len(lines[-1]) - len(lines[-1].lstrip())
            source_code += ' ' * num_spaces + '^' + '-' * (79 - num_spaces) + '\n'
        except:
            # Couldn't find source for some reason. Skip file.
            pass

    return (source_code, stack_trace, last_error_type)


